package ca.uwaterloo.crysp.attacktouchalytics;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;

import android.content.res.AssetManager;
import android.content.res.Resources;
import android.os.Environment;
import android.util.Log;

public class ModelStorage {
	static final String ds_prefix = "touchalytics_model_v1_";
	static final String out_prefix = "touchalytics_mimicry_attack_";
	
	public static int[] readModelStats(String user) {
		int [] arr = {0,0,0,0};
		/*String fileName = ds_prefix + user;
		File directory = Environment.getExternalStorageDirectory();
		File file = new File(directory + "/" + fileName);
		if (!file.exists()) {
			return arr;
		}*/
		BufferedReader reader = null;
		try {
			reader = new BufferedReader(
			        new InputStreamReader(MainActivity.mainActivity.getAssets().open(
			        		"raw/"+ds_prefix+user)));
			//reader = new BufferedReader(new FileReader(file));
			String line;
			while ((line = reader.readLine()) != null) {
				double [] fv = getFVFromString(line);
				arr[(int)fv[0]]++;
			}
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		return arr;
	}
	
	public static ArrayList<String> readModel(String user) {
		ArrayList <String> lines = new ArrayList<String>();
		/*String fileName = ds_prefix + user;
		File directory = Environment.getExternalStorageDirectory();
		File file = new File(directory + "/" + fileName);
		if (!file.exists()) {
			return lines;
		}*/
		BufferedReader reader = null;
		try {
			reader = new BufferedReader(
			        new InputStreamReader(MainActivity.mainActivity.getAssets().open(
			        		"raw/"+ds_prefix+user)));
			//reader = new BufferedReader(new FileReader(file));
			String line;
			while ((line = reader.readLine()) != null) {
				lines.add(line);
			}
		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		return lines;
	}

	public static void writeModel(String user, double [] fv, 
			ArrayList<TouchPoint> tps, ArrayList<String> sensorData) {
		//Log.e("line", deflateData(user, fv, tps, sensorData));
		
		String fileName = ds_prefix + user;
		File directory = Environment.getExternalStorageDirectory();
		File file = new File(directory + "/" + fileName);
		try {
			BufferedWriter out = 
					  new BufferedWriter(new FileWriter(file, true));
			out.write(deflateData(user, fv, tps, sensorData));
			//Log.e("Store", str.toString());
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
		} 
		
	}
	public static void writeToFile(String user, String line) {
		String fileName = out_prefix + user;
		File directory = Environment.getExternalStorageDirectory();
		File file = new File(directory + "/" + fileName);
		try {
			BufferedWriter out = 
					  new BufferedWriter(new FileWriter(file, true));
			out.write(line+"\n");
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
		} 
	}
	
	public static String getUserFromString(String str) {
		return str.substring(0, str.indexOf("|"));
	}
	
	public static double[] getFVFromString(String str) {
		String [] toks = str.split("\\|");
		String [] features = toks[1].split(",");
		double [] fv = new double[MainActivity.featureSetSize];
		for (int i = 0; i < MainActivity.featureSetSize; i++) 
			fv[i] = Double.parseDouble(features[i]);
		return fv;
	}
	
	public static ArrayList<TouchPoint> getTouchPointsFromString(
			String str) {
		String [] toks = str.split("\\|");
		ArrayList<TouchPoint> tps = new ArrayList<TouchPoint>();
		String [] points = toks[2].split(";"); 
		for (int i= 0; i < points.length; i++) {
			String [] p = points[i].split(",");
			long tstamp = Long.parseLong(p[0]);
			float x = Float.parseFloat(p[1]);
			float y = Float.parseFloat(p[2]);
			float area = Float.parseFloat(p[3]);
			TouchPoint tp = new TouchPoint(tstamp, x, y, area);
			tps.add(tp);
		}
		return tps;
	}
	public static String deflateData(String user, double [] fv, 
			ArrayList<TouchPoint> tps, ArrayList<String> sensorData) {
		StringBuilder str = new StringBuilder();
		str.append(user);
		str.append('|');
		for (int i = 0; i < fv.length; i++) {
			str.append(fv[i]);
			if (i < fv.length - 1)
				str.append(',');
		}
		str.append("|");
		for (int i = 0; i < tps.size(); i++) {
			str.append(tps.get(i).tstamp);
			str.append(',');
			str.append(tps.get(i).x);
			str.append(',');
			str.append(tps.get(i).y);
			str.append(',');
			str.append(tps.get(i).area);
			if (i < tps.size() - 1)
				str.append(';');
		}
		str.append("|");
		for (int i = 0; i < sensorData.size(); i++) {
			str.append(sensorData.get(i));
			if (i < sensorData.size() - 1)
				str.append(';');
		}
		str.append("\n");
		return str.toString();
	}
	//XXX: call writeModel for each fv
	public static void writeModelBatch(String user, ArrayList<double []> fv) {
		String fileName = ds_prefix + user;
		File directory = Environment.getExternalStorageDirectory();
		File file = new File(directory + "/" + fileName);
		try {
			BufferedWriter out = 
					  new BufferedWriter(new FileWriter(file, true));
			for (int i = 0; i < fv.size(); i++) {
				StringBuilder str = new StringBuilder();
				for (int j = 0; j < fv.get(i).length; j++) {
					str.append(fv.get(i)[j]);
					if (j < fv.get(i).length - 1)
						str.append(',');
				}
				str.append("\n");
				out.write(str.toString());
				Log.e("Store", str.toString());
			}
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
		} 
	}

	public static void deleteModel(String user) {
		String fileName = ds_prefix + user;
		File directory = Environment.getExternalStorageDirectory();
		String _fileName = directory + "/" + fileName;
		File file = new File(_fileName);
		if (file.exists())
			if(!file.delete())
				Log.e("Storage","Unable to delete the log file");
	}
	public static ArrayList<String> getModels() {
		Resources res = MainActivity.mainActivity.getResources(); //if you are in an activity
        AssetManager am = res.getAssets();
        String list[] = null;
		try {
			list = am.list("raw");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
        
		//File directory = new File("file:///android_asset/raw/");// Environment.getExternalStorageDirectory();
		//String [] list = directory.list();
		ArrayList<String> filteredList = new ArrayList<String>();
		for (int i = 0; i < list.length; i++) 
			if (list[i].startsWith(ds_prefix)) 
					filteredList.add(list[i].substring(ds_prefix.length()));
		return filteredList;
	}

}