package ca.uwaterloo.crysp.attacktouchalytics;

import java.util.ArrayList;

import android.util.Log;
import android.view.MotionEvent;

public class TouchFeatures {
	/*
	 * Swipe features used by Li et al., NDSS 2014
	 * 0- Direction: pivot for different training models
	 * 1,2- First touch position: x & y co-ords
	 * 3- First touch area: touch area of first touch co-ord
	 * 4- First moving direction: angle between horizontal line and this
	 * 5- Moving distance: sum of distance b/w consecutive points
	 * 6- Duration: time taken to swipe
	 * 7- Average moving direction: Avg val. of moving direction of n-1 pts
	 * 8- Average moving curvature: Avg val. of moving curvature of n-1 pts
	 * 					b/w consecutive pts ABC it is angle ABC 
	 * 9- Average touch area: Avg. of all touch areas
	 * 10- Max area: Index of Max-area pt./Total pts
	 */
	
	
	static double [] fv = new double[MainActivity.featureSetSize];
	
	static ArrayList<TouchPoint> points = new ArrayList<TouchPoint>();

	public static ArrayList<TouchPoint> getTouchPoints() {
		return points;
	}
	
	public static double[] getFV() {
		double [] _fv = new double[MainActivity.featureSetSize];
		for (int i = 0; i < fv.length; i++)
			_fv[i] = fv[i];
		return _fv;
	}
	private static void populateFeatureVector() {
		int numPoints = points.size();
		/*F 1-7 are pretty straightforward*/
		fv[0] = points.get(0).x; /*startX*/
		fv[1] = points.get(0).y;  /*startY*/
		fv[2] = points.get(numPoints - 1).x; /*endX*/
		fv[3] = points.get(numPoints - 1).y; /*endY*/
		fv[4] = points.get(numPoints - 1).tstamp - 
				points.get(0).tstamp; /*duration*/
		fv[5] = 0;
		//fv.set(5, this.touchPoints.get(numPoints - 1).eventTimestamp -
		//		TouchFeatures.lastSwipeTimestamp); /*interStrokeTime*/
		fv[6] = Math.sqrt(Math.pow(fv[2] - fv[0], 2) + 
				Math.pow(fv[3] - fv[1], 2)); /*directDistance*/

		/*Calculate pairwise displacement, velocity and acceleration*/
		double xDisplacement[] = new double[numPoints - 1];
		double yDisplacement[] = new double[numPoints - 1];
		double tDisplacement[] = new double[numPoints - 1];
		double pairwAngle[] = new double[numPoints - 1];
		double pairwDistance[] = new double[numPoints - 1];
		double pairwVelocity[] = new double[numPoints - 1];
		double pairwAcceleration[] = new double[numPoints - 2];

		for (int i = 0; i < numPoints - 2; i++) {
			xDisplacement[i] = points.get(i+1).x - 
					points.get(i).x;
			yDisplacement[i] = points.get(i+1).y - 
					points.get(i).y;
			tDisplacement[i] = points.get(i+1).tstamp - 
					points.get(i).tstamp;
			pairwAngle[i] =  Math.atan2(yDisplacement[i], xDisplacement[i]);
			pairwDistance[i] =  Math.sqrt(Math.pow(xDisplacement[i], 2) + 
					Math.pow(yDisplacement[i], 2));
			if (tDisplacement[i] == 0) 
				pairwVelocity[i] = 0;
			else 
				pairwVelocity[i] = pairwDistance[i]/tDisplacement[i];
		}
		/*correct pairwVelocity by setting '0' to maxVelocity*/
		double maxVelocity = ArrayUtil.max(pairwVelocity);
		for (int i = 0; i < pairwVelocity.length - 1; i++) 
			if (pairwVelocity[i] == 0)
				pairwVelocity[i] = maxVelocity;

		for (int i = 0; i < pairwVelocity.length - 2; i++) {
			pairwAcceleration[i] = pairwVelocity[i+1] - pairwVelocity[i];
			if (tDisplacement[i] == 0) 
				pairwAcceleration[i] = 0;
			else 
				pairwAcceleration[i] = pairwAcceleration[i]/tDisplacement[i];
		}
		/*calculate the max values for acceleration and replace
		 * values for which tDisplacement = 0 to max*/
		double maxAcceleration = 0;

		maxAcceleration = ArrayUtil.max(pairwAcceleration);

		for (int i = 0; i < pairwAcceleration.length - 1; i++) 
			if (pairwAcceleration[i] == 0)
				pairwAcceleration[i] = maxAcceleration;	

		/*F8-15*/
		fv[7] = ComplexNumbers.circ_r(pairwAngle); /*meanResultantLength*/
		fv[8] = ArrayUtil.percentile(pairwVelocity, 0.20); /*velocity20*/
		fv[9] = ArrayUtil.percentile(pairwVelocity, 0.50); /*velocity50*/
		fv[10] = ArrayUtil.percentile(pairwVelocity, 0.80); /*velocity80*/
		fv[11] = ArrayUtil.percentile(pairwAcceleration, 0.20); /*acceleration20*/
		fv[12] = ArrayUtil.percentile(pairwAcceleration, 0.50); /*acceleration50*/
		fv[13] = ArrayUtil.percentile(pairwAcceleration, 0.80); /*acceleration80*/
		fv[14] = Math.atan2(fv[3] - fv[1], fv[2] - fv[0]); /*lineDirection*/

		/*F16 last 3 velocity points*/
		double velocityPoints [] = {pairwVelocity[pairwVelocity.length - 1],
				pairwVelocity[pairwVelocity.length-2],
				pairwVelocity[pairwVelocity.length-3]};
		fv[15] = ArrayUtil.percentile(velocityPoints, 0.50); /*medVelocity*/

		/*F17-18: trajectoryLength & averageVelocity*/
		double temp = 0;

		for (int i = 0; i < pairwDistance.length; i++) {
			temp += pairwDistance[i]; /*trajectoryLength*/
		}
		fv[16] = temp;

		if(fv[4] == 0)
			fv[17] = 0;
		else
			fv[17] =  fv[16]/fv[4];

		/*F19 - First 5 acceleration points; medianAcceleration*/
		double accelerationPoints [] = {pairwAcceleration[0],
				pairwAcceleration[1], pairwAcceleration[2],
				pairwAcceleration[3],pairwAcceleration[4],
				pairwAcceleration[5]};
		fv[18] = ArrayUtil.percentile(accelerationPoints, 0.50);

		/*F20-22: midPressure, midArea, phoneOrientation*/
		fv[19] = 1;//touchPoints.get(numPoints/2).pressure;
		fv[20] = points.get(numPoints/2).area;
		fv[21] = 1;//touchPoints.get(numPoints/2).orientation;

		/*F23 - Direction Flag. up, down, left, right are 0,1,2,3*/
		fv[22] = 1;
		double xDiff = fv[2] - fv[0];
		double yDiff = fv[3] - fv[1];
		if (Math.abs(xDiff) > Math.abs(yDiff))
			if (xDiff < 0)
				fv[22] =  2; //left
			else
				fv[22] = 3; //right
		else
			if (yDiff < 0)
				fv[22] = 0; //up

		/*F24-25: distToTrajRatio; averageDirection*/
		if (fv[16] == 0)
			fv[23] = 0;
		else
			fv[23] = fv[6]/fv[16];

		fv[24] =ComplexNumbers.circ_mean(pairwAngle);

		/*F26-29 - Largest/20%/50%/80% deviation from end-to-end line*/
		double xVek [] = new double[numPoints];
		double yVek [] = new double[numPoints];
		for (int i = 0; i < numPoints - 1; i++) {
			xVek[i] = points.get(i).x - fv[0];
			yVek[i] = points.get(i).y - fv[1];
		}
		double perVek[] = {yVek[yVek.length-1], xVek[xVek.length-1] - 1, 0};
		temp = Math.sqrt(Math.pow(perVek[0], 2) + Math.pow(perVek[1], 2));
		if (temp == 0)
			perVek[0] = perVek[1] = perVek[2]  = 0;
		else {
			perVek[0] /= temp;
			perVek[1] /= temp;
			perVek[2] /= temp;
		}

		double absProj [] = new double[numPoints];
		for (int i = 0; i < numPoints - 1; i++) 
			absProj[i] = Math.abs(xVek[i] * perVek[0] + yVek[i] * perVek[1]);
		fv[25] = ArrayUtil.max(absProj);
		fv[26] = ArrayUtil.percentile(absProj, 0.2);
		fv[27] = ArrayUtil.percentile(absProj, 0.5);
		fv[28] = ArrayUtil.percentile(absProj, 0.8);

		
		//points.clear();

	}
	public static boolean procEvent(MotionEvent ev) {
    	switch (ev.getAction()) {
    	case MotionEvent.ACTION_DOWN:  /* primary pointer */
		case MotionEvent.ACTION_POINTER_DOWN: /* any subsequent pointer */
			points.clear();//clear points now; must've been fetched by now
			return false;
		case MotionEvent.ACTION_MOVE: /* any number of pointers move */
			for (int hIndx = 0; hIndx < ev.getHistorySize(); hIndx++) {
				for (int pIndex = 0; pIndex < ev.getPointerCount(); 
						pIndex++) {
					TouchPoint tp = new TouchPoint();
					tp.x = ev.getHistoricalX(pIndex, hIndx);
					tp.y = ev.getHistoricalY(pIndex, hIndx);
					tp.area = ev.getHistoricalSize(pIndex, hIndx);
					tp.tstamp = ev.getHistoricalEventTime(hIndx);
					points.add(tp);
				}
			}
			for (int pIndex = 0; pIndex < ev.getPointerCount(); 
					pIndex++) {
				TouchPoint tp = new TouchPoint();
				tp.x = ev.getX(pIndex);
				tp.y = ev.getY(pIndex);
				tp.area = ev.getSize(pIndex);
				tp.tstamp = ev.getEventTime();
				points.add(tp);
			}
			return false;
		case MotionEvent.ACTION_POINTER_UP: /* all pointers are up */
		case MotionEvent.ACTION_UP: 
		case MotionEvent.ACTION_CANCEL:
			if (points.size() > 10) {
				populateFeatureVector();
				return true;
			}
			//points.clear();
			return false;
    	}
    	return false;
	}

	static double[] TouchPointsToFV(ArrayList<TouchPoint> touchPoints) {
		double [] _fv = new double[MainActivity.featureSetSize];
		/*Feature List*/
		/**Start X (F1)*/
		/**Start Y (F2)*/
		/**End X (F3)*/
		/**End Y (F4)*/
		/**Duration in ms (F5)*/
		/**Inter-stroke time in ms (F6)*/
		/**Direct end-to-end distance (F7)*/
		/**Mean Resultant Length (F8)*/
		/**20% perc. pairwise velocity (F9)*/
		/**50% perc. pairwise velocity (F10)*/
		/**80% perc. pairwise velocity (F11)*/
		/**20% perc. pairwise acceleration (F12)*/
		/**50% perc. pairwise acceleration (F13)*/
		/**80% perc. pairwise acceleration (F14)*/
		/**Direction of End-to-End line (F15)*/
		/**Median velocity of last 3 points (F16)*/
		/**Length of Trajectory (F17)*/
		/**Average Velocity (F18)*/
		/**Median Acceleration at first 5 points (F19)*/
		/**Pressure in the middle of stroke (F20)*/
		/**Midstroke area covered (F21)*/
		/**Phone Orientation (F22)*/
		/**Direction Flag (F23)*/
		/**Ratio of Direct Distance and Traj. Length (F24)*/
		/**Average Direction of ensemble Pairs (F25)*/
		/**Largest Deviation from end-end Line (F26)*/
		/**20% perc. Deviation from end-to-end line (F27)*/
		/**50% perc. Deviation from end-to-end line (F28)*/
		/**80% perc. Deviation from end-to-end line (F29)*/

		int numPoints = touchPoints.size();
		/*F 1-7 are pretty straightforward*/
		_fv[0] = touchPoints.get(0).x; /*startX*/
		_fv[1] = touchPoints.get(0).y;  /*startY*/
		_fv[2] = touchPoints.get(numPoints - 1).x; /*endX*/
		_fv[3] = touchPoints.get(numPoints - 1).y; /*endY*/
		_fv[4] = touchPoints.get(numPoints - 1).tstamp - 
				touchPoints.get(0).tstamp; /*duration*/
		_fv[5] = 0;
		//fv.set(5, this.touchPoints.get(numPoints - 1).eventTimestamp -
		//		TouchFeatures.lastSwipeTimestamp); /*interStrokeTime*/
		_fv[6] = Math.sqrt(Math.pow(_fv[2] - _fv[0], 2) + 
				Math.pow(_fv[3] - _fv[1], 2)); /*directDistance*/

		/*Calculate pairwise displacement, velocity and acceleration*/
		double xDisplacement[] = new double[numPoints - 1];
		double yDisplacement[] = new double[numPoints - 1];
		double tDisplacement[] = new double[numPoints - 1];
		double pairwAngle[] = new double[numPoints - 1];
		double pairwDistance[] = new double[numPoints - 1];
		double pairwVelocity[] = new double[numPoints - 1];
		double pairwAcceleration[] = new double[numPoints - 2];

		for (int i = 0; i < numPoints - 2; i++) {
			xDisplacement[i] = touchPoints.get(i+1).x - 
					touchPoints.get(i).x;
			yDisplacement[i] = touchPoints.get(i+1).y - 
					touchPoints.get(i).y;
			tDisplacement[i] = touchPoints.get(i+1).tstamp - 
					touchPoints.get(i).tstamp;
			pairwAngle[i] =  Math.atan2(yDisplacement[i], xDisplacement[i]);
			pairwDistance[i] =  Math.sqrt(Math.pow(xDisplacement[i], 2) + 
					Math.pow(yDisplacement[i], 2));
			if (tDisplacement[i] == 0) 
				pairwVelocity[i] = 0;
			else 
				pairwVelocity[i] = pairwDistance[i]/tDisplacement[i];
		}
		/*correct pairwVelocity by setting '0' to maxVelocity*/
		double maxVelocity = ArrayUtil.max(pairwVelocity);
		for (int i = 0; i < pairwVelocity.length - 1; i++) 
			if (pairwVelocity[i] == 0)
				pairwVelocity[i] = maxVelocity;

		for (int i = 0; i < pairwVelocity.length - 2; i++) {
			pairwAcceleration[i] = pairwVelocity[i+1] - pairwVelocity[i];
			if (tDisplacement[i] == 0) 
				pairwAcceleration[i] = 0;
			else 
				pairwAcceleration[i] = pairwAcceleration[i]/tDisplacement[i];
		}
		/*calculate the max values for acceleration and replace
		 * values for which tDisplacement = 0 to max*/
		double maxAcceleration = 0;

		maxAcceleration = ArrayUtil.max(pairwAcceleration);

		for (int i = 0; i < pairwAcceleration.length - 1; i++) 
			if (pairwAcceleration[i] == 0)
				pairwAcceleration[i] = maxAcceleration;	

		/*F8-15*/
		_fv[7] = ComplexNumbers.circ_r(pairwAngle); /*meanResultantLength*/
		_fv[8] = ArrayUtil.percentile(pairwVelocity, 0.20); /*velocity20*/
		_fv[9] = ArrayUtil.percentile(pairwVelocity, 0.50); /*velocity50*/
		_fv[10] = ArrayUtil.percentile(pairwVelocity, 0.80); /*velocity80*/
		_fv[11] = ArrayUtil.percentile(pairwAcceleration, 0.20); /*acceleration20*/
		_fv[12] = ArrayUtil.percentile(pairwAcceleration, 0.50); /*acceleration50*/
		_fv[13] = ArrayUtil.percentile(pairwAcceleration, 0.80); /*acceleration80*/
		_fv[14] = Math.atan2(_fv[3] - _fv[1], _fv[2] - _fv[0]); /*lineDirection*/

		/*F16 last 3 velocity points*/
		double velocityPoints [] = {pairwVelocity[pairwVelocity.length - 1],
				pairwVelocity[pairwVelocity.length-2],
				pairwVelocity[pairwVelocity.length-3]};
		_fv[15] = ArrayUtil.percentile(velocityPoints, 0.50); /*medVelocity*/

		/*F17-18: trajectoryLength & averageVelocity*/
		double temp = 0;

		for (int i = 0; i < pairwDistance.length; i++) {
			temp += pairwDistance[i]; /*trajectoryLength*/
		}
		_fv[16] = temp;

		if(_fv[4] == 0)
			_fv[17] = 0;
		else
			_fv[17] =  _fv[16]/_fv[4];

		/*F19 - First 5 acceleration points; medianAcceleration*/
		double accelerationPoints [] = {pairwAcceleration[0],
				pairwAcceleration[1], pairwAcceleration[2],
				pairwAcceleration[3],pairwAcceleration[4],
				pairwAcceleration[5]};
		_fv[18] = ArrayUtil.percentile(accelerationPoints, 0.50);

		/*F20-22: midPressure, midArea, phoneOrientation*/
		_fv[19] = 1;//touchPoints.get(numPoints/2).pressure;
		_fv[20] = touchPoints.get(numPoints/2).area;
		_fv[21] = 1;//touchPoints.get(numPoints/2).orientation;

		/*F23 - Direction Flag. up, down, left, right are 0,1,2,3*/
		_fv[22] = 1;
		double xDiff = _fv[2] - _fv[0];
		double yDiff = _fv[3] - _fv[1];
		if (Math.abs(xDiff) > Math.abs(yDiff))
			if (xDiff < 0)
				_fv[22] =  2; //left
			else
				_fv[22] = 3; //right
		else
			if (yDiff < 0)
				_fv[22] = 0; //up

		/*F24-25: distToTrajRatio; averageDirection*/
		if (_fv[16] == 0)
			_fv[23] = 0;
		else
			_fv[23] = _fv[6]/_fv[16];

		_fv[24] = ComplexNumbers.circ_mean(pairwAngle);

		/*F26-29 - Largest/20%/50%/80% deviation from end-to-end line*/
		double xVek [] = new double[numPoints];
		double yVek [] = new double[numPoints];
		for (int i = 0; i < numPoints - 1; i++) {
			xVek[i] = touchPoints.get(i).x - _fv[0];
			yVek[i] = touchPoints.get(i).y - _fv[1];
		}
		double perVek[] = {yVek[yVek.length-1], xVek[xVek.length-1] - 1, 0};
		temp = Math.sqrt(Math.pow(perVek[0], 2) + Math.pow(perVek[1], 2));
		if (temp == 0)
			perVek[0] = perVek[1] = perVek[2]  = 0;
		else {
			perVek[0] /= temp;
			perVek[1] /= temp;
			perVek[2] /= temp;
		}

		double absProj [] = new double[numPoints];
		for (int i = 0; i < numPoints - 1; i++) 
			absProj[i] = Math.abs(xVek[i] * perVek[0] + yVek[i] * perVek[1]);
		_fv[25] = ArrayUtil.max(absProj);
		_fv[26] = ArrayUtil.percentile(absProj, 0.2);
		_fv[27] = ArrayUtil.percentile(absProj, 0.5);
		_fv[28] = ArrayUtil.percentile(absProj, 0.8);
		return _fv;

	}
}
